import logging
from django.db.models.fields.files import FieldFile, FileField, ImageField
from django.core.files.images import ImageFile

log = logging.getLogger(__name__)


class FileFieldMixin(object):
    def pre_save(self, model_instance, add):
        if getattr(self, "_storage_callable", None):
            if callable(self._storage_callable):
                self.storage = self._storage_callable(model_instance)
        return super().pre_save(model_instance, add)


class DynamicStorageFieldFile(FieldFile):
    """FieldFile implementation for dynamically setting the storage provider for the file at runtime based on a callable passed into the field definition"""

    def __init__(self, instance, field, name):
        if getattr(field, "_storage_callable", None):
            storage = field._storage_callable(instance)
            log.info("Setting storage for this field to {}".format(storage))
            field.storage = storage
        super().__init__(instance, field, name)


class DynamicStorageFileField(FileFieldMixin, FileField):
    """File field implementation for dynamically setting the storage provider for the file to be stored/retrieved at runtime based on a callable passed into this field definition for the `storage` kwarg"""

    attr_class = DynamicStorageFieldFile

    def pre_save(self, model_instance, add):
        if getattr(self, "_storage_callable", None):
            if callable(self._storage_callable):
                self.storage = self._storage_callable(model_instance)
        return super().pre_save(model_instance, add)


class DynamicStorageImageFieldFile(ImageFile, DynamicStorageFieldFile):
    def delete(self, save=True):
        if hasattr(self, "_dimensions_cache"):
            del self._dimensions_cache
        super().delete(save)


class DynamicStorageImageField(FileFieldMixin, ImageField):
    attr_class = DynamicStorageImageFieldFile



from django.test import TestCase
from .factories import (
    TestStorageTargetFactory,
    TestFileStorageModelFactory,
    TestEncryptedFileFieldModelFactory,
    TestEncryptedImageFieldModelFactory,
    TestImageStorageModelFactory,
)


class TestStorageTargetFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        self.assertIsNotNone(tstf.pk)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        self.assertIsNotNone(tstf.pk)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        self.assertIsNotNone(tstf.pk)


class TestFileStorageModelFactorytestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        tfsmf = TestFileStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNotNone(tfsmf.file.url)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        tfsmf = TestFileStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNotNone(tfsmf.file.url)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        tfsmf = TestFileStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNotNone(tfsmf.file.url)


class TestImageStorageModelFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        tismf = TestImageStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNotNone(tismf.image.url)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        tismf = TestImageStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNotNone(tismf.image.url)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        tismf = TestImageStorageModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNotNone(tismf.image.url)


class TestEncryptedFileFieldModelFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        tfsmf = TestEncryptedFileFieldModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNone(tfsmf.file.url)
        self.assertIsNotNone(tfsmf.file)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        tfsmf = TestEncryptedFileFieldModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNone(tfsmf.file.url)
        self.assertIsNotNone(tfsmf.file)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        tfsmf = TestEncryptedFileFieldModelFactory(storage_target=tstf)
        self.assertIsNotNone(tfsmf.pk)
        self.assertIsNone(tfsmf.file.url)
        self.assertIsNotNone(tfsmf.file)


class TestEncryptedImageFieldModelFactoryTestCase(TestCase):
    def test_gcs(self):
        tstf = TestStorageTargetFactory(provider="gcloud")
        tismf = TestEncryptedImageFieldModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNone(tismf.image.url)
        self.assertIsNotNone(tismf.image)

    def test_aws(self):
        tstf = TestStorageTargetFactory(provider="s3boto3")
        tismf = TestEncryptedImageFieldModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNone(tismf.image.url)
        self.assertIsNotNone(tismf.image)

    def test_azure(self):
        tstf = TestStorageTargetFactory(provider="azure")
        tismf = TestEncryptedImageFieldModelFactory(storage_target=tstf)
        self.assertIsNotNone(tismf.pk)
        self.assertIsNone(tismf.image.url)
        self.assertIsNotNone(tismf.image)



from .models import TestStorageTarget, TestFileStorageModel, TestImageStorageModel, TestEncryptedFileFieldModel, TestEncryptedImageFieldModel, get_fernet, gen_key
import base64
import json
import os
import string
from datetime import timedelta
from dynamic_storages.models.mappings import LAST_STATUS_CHOICES, STORAGE_PROVIDER_MAP
import factory
import factory.django
import factory.fuzzy
from django.utils import timezone
from django.utils.text import slugify
from faker import Faker
from functools import partial
from django.core.files.base import ContentFile
import logging


log = logging.getLogger(__name__)

fake = Faker()


def get_config(key, **kwargs):
    log.debug("Getting config for {}".format(key))
    ret_val = {}
    if key.lower() == "alc":
        pass
    elif key.lower() in ["as", "azure"]:
        ret_val = {
            "account_key": kwargs.get(
                "account_key",
                os.getenv(
                    "TEST_AZURE_ACCOUNT_KEY",
                ),
            ),
            "account_name": kwargs.get(
                "account_name",
                os.getenv("TEST_AZURE_ACCOUNT_NAME"),
            ),
            "overwrite_files": bool(kwargs.get("overwrite_files", "True")),
            "azure_container": kwargs.get(
                "azure_container",
                os.getenv("TEST_AZURE_CONTAINER"),
            ),
        }
    elif key.lower() == "dropbox":
        ret_val = {
            "oauth2_access_token": kwargs.get("oauth2_access_token", os.getenv("TEST_DBS_OAUTH2_TOKEN")),
            "root_path": kwargs.get("root_path", os.getenv("TEST_DBS_ROOT_PATH")),
        }
    elif key.lower() == "ftp":
        ret_val = {
            "location": "ftp://{}:{}@{}:{}/".format(
                kwargs.get("username", os.getenv("TEST_FTP_USER_NAME")),
                kwargs.get("password", os.getenv("TEST_FTP_USER_PASS")),
                kwargs.get("host", os.getenv("TEST_FTP_HOST")),
                kwargs.get("port", os.getenv("TEST_FTP_PORT")),
            )
        }
    elif key.lower() == "gcloud":
        ret_val = {
            "bucket_name": kwargs.get(
                "bucket_name",
                os.getenv("TEST_GCS_BUCKET_NAME"),
            ),
            "credentials": json.loads(base64.b64decode(os.getenv("TEST_GCS_SERVICE_ACCOUNT_JSON", "{}"))),
        }
    elif key.lower() == "do":
        ret_val = {
            "access_key": kwargs.get("access_key", os.getenv("TEST_DO_ACCESS_KEY")),
            "secret_key": kwargs.get("secret_key", os.getenv("TEST_DO_SECRET_KEY")),
            "bucket_name": kwargs.get(
                "bucket_name",
                os.getenv("TEST_DO_BUCKET_NAME"),
            ),
            "region_name": kwargs.get("region_name", os.getenv("TEST_DO_REGION_NAME", "sfo2")),
            "default_acl": kwargs.get("default_acl", "private"),
        }
    elif key.lower() in ["s3b", "aws", "s3boto3"]:
        ret_val = {
            "access_key": kwargs.get("access_key", os.getenv("TEST_AWS_ACCESS_KEY")),
            "secret_key": kwargs.get("secret_key", os.getenv("TEST_AWS_SECRET_KEY")),
            "file_overwrite": kwargs.get("file_overwrite", True),
            "bucket_name": kwargs.get(
                "bucket_name",
                os.getenv("TEST_AWS_BUCKET_NAME", "introsepect-data-django-test"),
            ),
            "region_name": kwargs.get("region_name", os.getenv("TEST_AWS_REGION_NAME")),
            "default_acl": kwargs.get("default_acl", "private"),
        }
    elif key.lower() == "sftp":
        SSH_KEY_PATH = os.path.join(TEST_DATA_BASE_DIR, "sftp")
        with open(os.path.join(SSH_KEY_PATH, "sftp-test"), "r") as f:
            ssh_private_key = f.read()

        ret_val = {
            "host": kwargs.get("host", os.getenv("TEST_SFTP_HOST")),
            "params": {
                "port": int(kwargs.get("port", os.getenv("TEST_SFTP_PORT"))),
                "username": kwargs.get("username", os.getenv("TEST_SFTP_USER_NAME")),
                "password": kwargs.get("password", os.getenv("TEST_SFTP_USER_PASS")),
                "pkey": ssh_private_key,
            },
            "interactive": None,
            "file_mode": None,
            "dir_mode": None,
            "uid": None,
            "gid": None,
            "known_host_file": None,
            "root_path": None,
            "base_url": None,
        }
    log.debug("Config for {} is {}".format(key, ret_val))
    return ret_val


class TestStorageTargetFactory(factory.django.DjangoModelFactory):
    name = factory.LazyFunction(fake.catch_phrase)
    description = factory.LazyFunction(partial(fake.paragraph, 3))
    provider = factory.fuzzy.FuzzyChoice([x for x in STORAGE_PROVIDER_MAP.keys()], getter=lambda c: c)
    config = factory.LazyAttribute(lambda o: get_config(o.provider))
    last_status = factory.fuzzy.FuzzyChoice([x for x in LAST_STATUS_CHOICES], getter=lambda c: c[0])

    class Meta:
        model = TestStorageTarget


def gen_file():
    return ContentFile(fake.binary(), name=fake.file_name())


class TestFileStorageModelFactory(factory.django.DjangoModelFactory):
    storage_target = factory.SubFactory(TestStorageTargetFactory)
    file = factory.django.FileField(from_func=gen_file)

    class Meta:
        model = TestFileStorageModel


class TestImageStorageModelFactory(factory.django.DjangoModelFactory):
    storage_target = factory.SubFactory(TestStorageTargetFactory)
    image = factory.django.ImageField()

    class Meta:
        model = TestImageStorageModel


class TestEncryptedFileFieldModelFactory(factory.django.DjangoModelFactory):
    storage_target = factory.SubFactory(TestStorageTargetFactory)
    file = factory.django.FileField(from_func=gen_file)

    class Meta:
        model = TestEncryptedFileFieldModel


class TestEncryptedImageFieldModelFactory(factory.django.DjangoModelFactory):
    storage_target = factory.SubFactory(TestStorageTargetFactory)
    image = factory.django.ImageField()

    class Meta:
        model = TestEncryptedImageFieldModel



from django.views.generic import View
import os
import magic
from .conf import settings
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404


class AbstractSecureFileContents(View):
    model = None
    url_kwargs = []
    file_field = None

    def get(self, request, *args, **kwargs):
        user_id = request.user
        try:
            query_kwargs = {k: kwargs.get(k) for k in url_kwargs if kwargs.get(k)}
            if request.user.is_authenticated:
                doc = get_object_or_404(self.model, **query_kwargs)
                content = getattr(doc, file_field).get_decrypted().read()
                return HttpResponse(content, content_type=magic.Magic(mime=True).from_buffer(content))
            else:
                log.warn("Unauthenticated request for {} with kwargs: {}".format(self.model.__name__, query_kwargs))
        except Exception as ex:
            log.warn("Failed to get property for request: {}".format(ex))
        raise Http404



from dynamic_storages.models import AbstractStorageTarget
from django.db import models
from dynamic_storages.fields.dynamic_storage import DynamicStorageFileField, DynamicStorageImageField
from dynamic_storages.fields.encrypted_content import EncryptedFileField, EncryptedImageField
from django.core.files.storage import default_storage
from cryptography.fernet import Fernet
import base64
import logging
import hashlib
from uuid import uuid4
from dynamic_storages.conf import settings

log = logging.getLogger(__name__)


class TestStorageTarget(AbstractStorageTarget):
    class Meta(AbstractStorageTarget.Meta):
        abstract = False


class TestBase(models.Model):
    storage_target = models.ForeignKey("TestStorageTarget", on_delete=models.CASCADE)

    class Meta:
        abstract = True


def get_storage(instance=None):
    if instance:
        return instance.storage_target.storage_backend
    return default_storage


def upload_to(instance, filename):
    prefix = hashlib.md5(str(uuid4()).encode("utf-8")).hexdigest()
    return "{}/{}".format(prefix, filename)


class TestFileStorageModel(TestBase):
    file = DynamicStorageFileField(storage=get_storage, upload_to=upload_to)

    class Meta:
        abstract = False


class TestImageStorageModel(TestBase):
    image = DynamicStorageImageField(storage=get_storage, upload_to=upload_to)

    class Meta:
        abstract = False


def get_fernet(instance, prop_name="key"):
    if not getattr(instance, prop_name, None):
        print("Setting key for fernet in instance")
        setattr(instance, prop_name, gen_key())
    return Fernet(base64.b64decode(getattr(instance, prop_name, None)))


def gen_key():
    return base64.b64encode(Fernet.generate_key())


class TestEncryptedFileFieldModel(TestBase):
    file = EncryptedFileField(storage=get_storage, fernet=get_fernet, upload_to=upload_to)
    key = models.CharField(max_length=60, default=gen_key, editable=False)

    class Meta:
        abstract = False


class TestEncryptedImageFieldModel(TestBase):
    image = EncryptedImageField(storage=get_storage, fernet=get_fernet, upload_to=upload_to)
    key = models.CharField(max_length=60, default=gen_key, editable=False)

    class Meta:
        abstract = False
